//
// This file contains model number definitions for the cups-filters sample
// ESC/P driver.
//
// Copyright 2007 by Apple Inc.
// Copyright 1997-2005 by Easy Software Products.
//
// Licensed under Apache License v2.0.  See the file "LICENSE" for more
// information.
//

#define EPSON_9PIN	0		// 9-pin dot matrix
#define EPSON_24PIN	1		// 24-pin dot matrix
#define EPSON_COLOR	2		// Epson Stylus Color with ESC .
#define EPSON_PHOTO	3		// Epson Stylus Photo with ESC .
#define EPSON_ICOLOR	4		// Epson Stylus Color with ESC i
#define EPSON_IPHOTO	5		// Epson Stylus Photo with ESC i
